// IPCReadWriteLock.cpp: implementation of the CIPCReadWriteLock class.
//
//////////////////////////////////////////////////////////////////////

// The Single Writer/Multiple Reader Guard
// for inter-process communication.
// Written by: Alex Farber   alexm@cmt.co.il  
// 2001
//
// Based on CSWMRG class 
// Copyright (c) 2000 Jeffrey Richter
// From the book:
// Programming Applications for Microsoft Windows
// PUBLISHED BY
// Microsoft Press


#include "stdafx.h"
#include "IPCReadWriteLock.h"


#define c_nLen 6
#define c_sMutexName TEXT("_Mutex")
#define c_sSemReader TEXT("_SemRd")
#define c_sSemWriter TEXT("_SemWr")
#define c_sMemFile   TEXT("_MemFl")


#ifdef _DEBUG
#define RWL_TRACE(x) OutputDebugString(x)
#else
#define RWL_TRACE(x)
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CIPCReadWriteLock::CIPCReadWriteLock(
             LPCTSTR sName,        // [in] name
             DWORD dwMilliseconds) // [in] maximum number of milliseconds to wait
                                   //      or INFINITE
{
    // Generate names for mutex and semaphores.
    // Names start from sName string and finish with
    // _Mutex, _SemRd and _SemWr.
    // If number of instances of this class are created
    // in the same process with the same name (also NULL),
    // they will guard the same resource.

    int n = 0;

    if ( sName )
        n = lstrlen(sName);

    n += c_nLen + 1;

    m_sMutexName = new TCHAR[n];
    m_sSemReadersName = new TCHAR[n];
    m_sSemWritersName = new TCHAR[n];
    m_sMemFileName = new TCHAR[n];

    *m_sMutexName = *m_sSemReadersName = *m_sSemWritersName = 0;

    if ( sName )
    {
        lstrcpy(m_sMutexName, sName);
        lstrcpy(m_sSemReadersName, sName);
        lstrcpy(m_sSemWritersName, sName);
        lstrcpy(m_sMemFileName, sName);
    }

    lstrcat(m_sMutexName, c_sMutexName);
    lstrcat(m_sSemReadersName, c_sSemReader);
    lstrcat(m_sSemWritersName, c_sSemWriter);
    lstrcat(m_sMemFileName, c_sMemFile);

    // keep time interval
    m_dwMilliseconds = dwMilliseconds;

    m_bInitSucceded = InitSyncObjects();

}

// Initialize syncronization values.
// Return value: TRUE - OK
//               FALSE - failed
//
// If this function returns FALSE, every call to
// WaitToRead, WaitToWrite, Done fails.
BOOL CIPCReadWriteLock::InitSyncObjects()
{
    // create semaphore for waiting readers
    m_hsemReaders = CreateSemaphore(NULL,              // security attributes
        0,                 // initial count
        MAXLONG,           // maximum count
        m_sSemReadersName);// name
    
    if ( ! m_hsemReaders )
    {
        RWL_TRACE(TEXT("Create readers semaphore failed\n"));
        return FALSE;
    }
    
    // create semaphore for waiting writers
    m_hsemWriters = CreateSemaphore(NULL, 
        0, 
        MAXLONG, 
        m_sSemWritersName);
    
    if ( ! m_hsemWriters )
    {
        RWL_TRACE(TEXT("Create writers semaphore failed\n"));
        return FALSE;
    }

    // crate mutex protecting access to other members
    m_hMutex = CreateMutex(NULL,             // security attributes
        FALSE,            // initial owner
        m_sMutexName);    // name
    
    
    if ( ! m_hMutex )
    {
        RWL_TRACE(TEXT("CreateMutex failed\n"));
        return FALSE;
    }

    // create file mapping to keep counters
    m_hFileMapping = CreateFileMapping(
        INVALID_HANDLE_VALUE,           // system paging file
        NULL,                           // security attributes
        PAGE_READWRITE,                 // protection
        0,                              // high-order DWORD of size
        sizeof(RW_COUNTERS),            // low-order DWORD of size
        m_sMemFileName);                // name
    
    if ( ! m_hFileMapping )
    {
        RWL_TRACE(TEXT("CreateFileMapping failed\n"));
        return FALSE;
    }
    
    m_pViewOfFile = MapViewOfFile(
        m_hFileMapping,             // handle to file-mapping object
        FILE_MAP_ALL_ACCESS,        // desired access
        0,
        0,
        0);                         // map all file
    
    if ( ! m_pViewOfFile )
    {
        RWL_TRACE(TEXT("MapViewOfFile failed\n"));
        return FALSE;
    }
 
    // get pointers to counters
    RW_COUNTERS* pCounters = (RW_COUNTERS*) m_pViewOfFile;
    
    m_pnActive = &pCounters->m_nActive;
    m_pnWaitingReaders = &pCounters->m_nWaitingReaders;
    m_pnWaitingWriters = &pCounters->m_nWaitingWriters;
    
    // If memory-mapped file is created, all pages are zero,
    // so I don't care about setting start value of counters.
    
    return TRUE;
}

// clean-up operations
CIPCReadWriteLock::~CIPCReadWriteLock()
{
    delete[] m_sMutexName;
    delete[] m_sSemReadersName;
    delete[] m_sSemWritersName;
    delete[] m_sMemFileName;

    if ( m_hFileMapping )
    {
        if ( m_pViewOfFile )
        {
            UnmapViewOfFile(m_pViewOfFile);
        }

        CloseHandle(m_hFileMapping);
    }

    if ( m_hMutex )
    {
        CloseHandle(m_hMutex);
    }

    if ( m_hsemWriters )
    {
        CloseHandle(m_hsemWriters);
    }

    if ( m_hsemReaders )
    {
        CloseHandle(m_hsemReaders);
    }

}


// Get read access
// Return value: TRUE - OK
//               FALSE - failed
BOOL CIPCReadWriteLock::WaitToRead() 
{
    if ( ! m_bInitSucceded )
        return FALSE;


    // Ensure exclusive access to the member variables
    if ( WaitForSingleObject(m_hMutex, m_dwMilliseconds) != WAIT_OBJECT_0 )
        return FALSE;

    
    // Are there writers waiting or is a writer writing?
    BOOL fResourceWritePending = (*m_pnWaitingWriters || (*m_pnActive < 0));
    
    if (fResourceWritePending) 
    {
        // This reader must wait, increment the count of waiting readers
        (*m_pnWaitingReaders)++;
    } 
    else 
    {
        // This reader can read, increment the count of active readers
        (*m_pnActive)++;
    }
    
    // Allow other clients to attempt reading/writing
    ReleaseMutex(m_hMutex);
    
    if ( fResourceWritePending ) 
    {
        // This thread must wait
        if ( WaitForSingleObject(m_hsemReaders, 
                                 m_dwMilliseconds) != WAIT_OBJECT_0 )
        {
            // try to make rollback
            if ( WaitForSingleObject(m_hMutex, 
                                     m_dwMilliseconds) == WAIT_OBJECT_0 )
            {
                (*m_pnWaitingReaders)--;    // rollback

                ReleaseMutex(m_hMutex);
            }

            return FALSE;
        }
    }

    return TRUE;
}


// Get write access
// Return value: TRUE - OK
//               FALSE - failed
BOOL CIPCReadWriteLock::WaitToWrite() 
{
    if ( ! m_bInitSucceded )
        return FALSE;

    // Ensure exclusive access to the member variables
    if ( WaitForSingleObject(m_hMutex, m_dwMilliseconds) != WAIT_OBJECT_0 )
        return FALSE;
    
    // Are there any threads accessing the resource?
    BOOL fResourceOwned = (*m_pnActive != 0);
    
    if (fResourceOwned) 
    {
        // This writer must wait, increment the count of waiting writers
        (*m_pnWaitingWriters)++;
    } 
    else 
    {
        // This writer can write, decrement the count of active writers
        *m_pnActive = -1;
    }
    
    // Allow other clients to attempt reading/writing
    ReleaseMutex(m_hMutex);
    
    if ( fResourceOwned ) 
    {
        // This thread must wait
        if ( WaitForSingleObject(m_hsemWriters, 
                                 m_dwMilliseconds) != WAIT_OBJECT_0 )
        {
            // try to make rollback
            if ( WaitForSingleObject(m_hMutex, 
                                     m_dwMilliseconds) == WAIT_OBJECT_0 )
            {
                (*m_pnWaitingWriters)--;        // rollback

                ReleaseMutex(m_hMutex);
            }

            return FALSE;
        }
    }

    return TRUE;
}


// Release common resource
// Return value: TRUE - OK
//               FALSE - failed
BOOL CIPCReadWriteLock::Done() 
{
    if ( ! m_bInitSucceded )
        return FALSE;
    
    // Ensure exclusive access to the member variables
    if ( WaitForSingleObject(m_hMutex, m_dwMilliseconds) != WAIT_OBJECT_0 )
        return FALSE;
    
    if (*m_pnActive > 0) 
    {
        // Readers have control so a reader must be done
        (*m_pnActive)--;
    } 
    else 
    {
        // Writers have control so a writer must be done
        (*m_pnActive)++;
    }
    
    HANDLE hsem = NULL;  // Assume no threads are waiting
    LONG lCount = 1;     // Assume only 1 waiter wakes; always true for writers
    
    if ( *m_pnActive == 0 ) 
    {
        // No thread has access, who should wake up?
        // NOTE: It is possible that readers could never get access
        //       if there are always writers wanting to write
        
        if ( *m_pnWaitingWriters > 0 ) 
        {
            // Writers are waiting and they take priority over readers
            *m_pnActive = -1;         // A writer will get access
            (*m_pnWaitingWriters)--;  // One less writer will be waiting
            hsem = m_hsemWriters;   // Writers wait on this semaphore
            // NOTE: The semaphore will release only 1 writer thread
        } 
        else if ( *m_pnWaitingReaders > 0 ) 
        {
            // Readers are waiting and no writers are waiting
            *m_pnActive = *m_pnWaitingReaders;   // All readers will get access
            *m_pnWaitingReaders = 0;             // No readers will be waiting
            hsem = m_hsemReaders;                // Readers wait on this semaphore
            lCount = *m_pnActive;                // Semaphore releases all readers
        } 
        else 
        {
            // There are no threads waiting at all; no semaphore gets released
        }
    }
    
    // Allow other threads to attempt reading/writing
    ReleaseMutex(m_hMutex);
    
    if (hsem != NULL) 
    {
        // Some threads are to be released
        ReleaseSemaphore(hsem, lCount, NULL);
    }

    return TRUE;
}

